%
% This small demo compares MULTIpLE algorithm to the Regularized Least-Squares
% on the five classes from the Caltech-256 dataset.
%
% The algorithm is described in
%
% "From N to N+ 1: Multiclass Transfer Incremental Learning."
% Kuzborskij, Ilja, Francesco Orabona, and Barbara Caputo.
% Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference on. IEEE, 2013.
%
% Contact the author: ilja.kuzborskij [at] idiap.ch
%                     http://idiap.ch/~ikuzbor
%
% ----------------------------------------------------------------------------------
%
% Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
% Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
%
% This file is part of MULTIpLE.
%
% MULTIpLE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License version 3 as
% published by the Free Software Foundation.
%
% MULTIpLE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with MULTIpLE. If not, see <http://www.gnu.org/licenses/>.
%

% Loading data
train = load('data/train.mat');
% Loading source prediction for these data
src_pred_train = load('data/src_pred_train_linear.mat');

% Hyperparameters are fixed
hp.lambda = 1;
hp.tau = 1;

% Will evaluate over 10 data splits and on progression of m examples per class
splits = 10;
train_m = [2, 5, 10, 15];

multiple_scores = [];
rls_scores = [];

classes = unique(train.y);
for s = 1:splits % Data splits
    fprintf('MULTIpLE demo:: Evaluating data split %d/%d...\n', s, splits);
    for i = 1:numel(train_m) % Number of examples per class
        for j = 1:numel(train.classes) % Leave-one-class-out evaluation
            % Holding out the test sample
            part = cvpartition(train.y, 'holdout', size(train.x, 1) - train_m(i)*numel(classes));

            % Computing linear kernels
            K_train = train.x(part.training, :) * train.x(part.training, :)';
            K_test = train.x(part.test, :) * train.x(part.training, :)';

            train_y = train.y(part.training);
            test_y = train.y(part.test);

            % Leaving-out predictions of source for target class
            src_pred_train_leave_c = src_pred_train.x(part.training, classes ~= classes(j) );
            src_pred_test_leave_c = src_pred_train.x(part.test, classes ~= classes(j));

            % Training and evaluating MULTIpLE
            [model, ~, yloo] = train_multiple(hp, K_train, src_pred_train_leave_c, train_y, classes(j));
            multiple_scores(s,i,j) = eval_multiple(model, K_test, src_pred_test_leave_c, test_y);

            % Training and evaluating regularized least squares (the baseline)
            preds_rls = K_test * ((K_train + hp.lambda*eye(size(K_train))) \  ova(train_y));
            [~, ova_code] = ova(test_y);
            [~,ix] = max(preds_rls, [], 2);
            rls_scores(s,i,j) = mean(ova_code(ix) == test_y);
        end
    end
end

% Plotting results
hold on;
plot(train_m, mean(mean(multiple_scores, 3),1), 'r', 'linewidth', 3);
plot(train_m, mean(mean(rls_scores, 3),1), 'b', 'linewidth', 3);

legend({'MULTIpLE', 'Regularized Least Squares'}, 'location', 'SouthEast', 'fontweight','bold','fontsize',12);
title('Recognition Accuracy: MULTIpLE vs. RLS (5 unrelated classes)', 'fontweight','bold','fontsize',12);
xlabel('Number of examples per class', 'fontweight','bold','fontsize',12);
ylabel('Recognition Accuracy', 'fontweight','bold','fontsize',12);

