%
% This function evaluates the trained MULTIpLE model given the kernel matrix
% between the training and testing samples, source classifier predictions on the
% testing set and test labels.
%
% [acc, pred, scores] = eval_multiple(model, K, src_pred, y)
%
% Input:
% model -- model structure generated by train_multiple function
% K -- kernel matrix between the testing and training examples
% src_pred -- a matrix containing predicitons of the source classifier on the testing data (#examples x #source-classes)
% K -- gram matrix of training data of size m
% src_pred -- source predictions on training data (# examples X # sources)
% y -- testing labels

%
% Output:
% acc -- recognition accuracy w.r.t. 0/1 loss function
% pred -- predicted labels
% scores -- scores generated by the predictor (margins)
%
% The MULTIpLE algorithm is described in
%
% "From N to N+ 1: Multiclass Transfer Incremental Learning."
% Kuzborskij, Ilja, Francesco Orabona, and Barbara Caputo.
% Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference on. IEEE, 2013.
%
% Contact the author: ilja.kuzborskij [at] idiap.ch
%                     http://idiap.ch/~ikuzbor
%
% ----------------------------------------------------------------------------------
%
% Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
% Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
%
% This file is part of MULTIpLE.
%
% MULTIpLE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License version 3 as
% published by the Free Software Foundation.
%
% MULTIpLE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with MULTIpLE. If not, see <http://www.gnu.org/licenses/>.
%

function [acc, pred, scores] = eval_multiple(model, K, src_pred, y)
    m = size(src_pred, 1);
    N = numel(unique(y));
    src_scores = zeros(m, N);
    src_scores(:, ~model.new_class_ix) = src_pred;
    src_scores(:, model.new_class_ix) = src_pred * model.beta;

    scores = bsxfun(@plus, K * model.A, model.b) + src_scores;

    [~, ix] = max(scores, [], 2);
    pred = model.ova_code(ix);

    acc = mean(pred == y);
end
