/*
 Trains a target classfier for N+1 classes given the predictions of the source classifier for the N classes and the training set
 containing examples from all the N+1 classes. The new, N+1-th class is trained by leveraging upon the provided N class classifier predictions.

 MultipleModel train_multiple(MatrixXd &K, MatrixXd &Y, MatrixXd &src_pred, size_t new_class_pos, size_t iterations, double lambda, double tau);

 Input:
 K -- training data gram matrix
 Y -- training label matrix by One-Versus-All coding (# examples X # classes)
 src_pred -- source predictions on training data (# examples X # sources)
 new_class_pos -- position of the new class amongst Y columns
 iterations -- number of iterations (subgradient descent)
 lambda -- regularization parameter of RLS classifier with biased regularization
 tau -- regularization parameter for the transfer coefficients (projection radius for subgradient descent)

 Output:
 MultipleModel structure instance containing:
   A -- dual (KRLS) parameters
   b -- bias parameter
   beta -- transfer parameters
   yloo -- leave-one-out error matrix for the last subgradient descent iteration

 The algorithm is described in

 "From N to N+ 1: Multiclass Transfer Incremental Learning."
 Kuzborskij, Ilja, Francesco Orabona, and Barbara Caputo.
 Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference on. IEEE, 2013.

 Contact the author: ilja.kuzborskij [at] idiap.ch
                     http://idiap.ch/~ikuzbor

 ----------------------------------------------------------------------------------

 Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
 Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>

 This file is part of MULTIpLE.

 MULTIpLE is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 3 as
 published by the Free Software Foundation.

 MULTIpLE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with MULTIpLE. If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <stdexcept>
#include <limits>
#include <Eigen/Core>
#include <Eigen/Dense>

using namespace Eigen;
using namespace std;

struct MultipleModel {
    MatrixXd A;
    MatrixXd b;
    VectorXd beta;
    MatrixXd yloo;
};

MultipleModel train_multiple(MatrixXd &K, MatrixXd &Y, MatrixXd &src_pred, size_t new_class_pos, size_t iterations, double lambda, double tau) {
    size_t m = Y.rows(); // Examples
    size_t k = Y.cols(); // Classes
    size_t n = src_pred.cols(); // Sources

    if (n != k-1)
        throw logic_error("Number of sources should be by one less than number of classes.");

    MatrixXd M(m+1, m+1);
    M << K + lambda*MatrixXd::Identity(m,m), MatrixXd::Ones(m,1),
        MatrixXd::Ones(1,m), 0;
    M = M.inverse();

    DiagonalMatrix<double,Dynamic> MoI(VectorXd(M.topLeftCorner(m,m).diagonal()));
    MoI = MoI.inverse();

    MatrixXd tmp(m+1, k);
    tmp << Y, MatrixXd::Zero(1, k);
    tmp = M * tmp;
    MatrixXd b_ = tmp.bottomRows(1);
    MatrixXd A_ = tmp.topRows(m);

    tmp = MatrixXd::Zero(m+1, n);

    tmp << src_pred, MatrixXd::Zero(1, n);
    tmp = M * tmp;
    MatrixXd b__ = tmp.bottomRows(1);
    MatrixXd A__ = tmp.topRows(m);

    MatrixXd A_Abeta(m, k);
    A_Abeta << A__.block(0, 0, m, new_class_pos),
        MatrixXd::Zero(m, 1),
        A__.block(0, new_class_pos, m, n-new_class_pos);

    MatrixXd yloo(m, n+1);
    VectorXd subgrad(n, 1);
    double total_loss = 0;

    VectorXd beta(n, 1);
    beta = VectorXd::Zero(n, 1);

    // Subgradient descent
    for (size_t t = 0; t < iterations; ++t) {
        A_Abeta.col(new_class_pos) = A__ * beta;
        yloo = Y - MoI * (A_ - A_Abeta);

        subgrad = VectorXd::Zero(n, 1);
        total_loss = 0;

        // Computing subgradient descent w.r.t. beta
        for (size_t i = 0; i < m; ++i) {
            VectorXd i_class_mask( (Y.row(i)+MatrixXd::Ones(1,k))/2 );
            VectorXd not_i_class_mask( (-Y.row(i)+MatrixXd::Ones(1,k))/2 );

            // Modified Crammer-Singer loss function
            if (Y(i, new_class_pos) != 1) { // If example belongs to the new class
                double loss = 1 + yloo(i, new_class_pos) - (yloo.row(i) * i_class_mask);
                if (loss > 0) {
                    subgrad += (A__.row(i) / M(i,i));
                    total_loss += loss;
                }
            }
            else {
                double yloo_curr_sample = (yloo.row(i) * i_class_mask);
                VectorXd yloo_not_curr_sample = not_i_class_mask.transpose().cwiseProduct(yloo.row(i));
                yloo_not_curr_sample -= i_class_mask * numeric_limits<double>::max(); // a trick, so that curr sample index is never selected

                double loss = (VectorXd::Ones(k) + yloo_not_curr_sample - yloo_curr_sample * VectorXd::Ones(k)).maxCoeff();
                if (loss > 0) {
                    subgrad -= (A__.row(i) / M(i,i));
                    total_loss += loss;
                }
            }
        }

        beta -= subgrad / (m * sqrt(double(t+1)));

        // Projection
        beta = (beta.array() < 0).select(0, beta);
        double beta_norm = beta.norm();
        if (beta_norm > tau) {
            beta /= beta_norm;
        }
    }

    MultipleModel model;

    // Finalizing by solving w.r.t. "gap" parameters
    A_Abeta.col(new_class_pos) = A__ * beta;
    model.A = A_ - A_Abeta;

    MatrixXd b_bbeta(1, k);
    b_bbeta << b__.block(0, 0, 1, new_class_pos),
        b__ * beta,
        b__.block(0, new_class_pos, 1, n-new_class_pos);
    model.b = b_ - b_bbeta;

    model.beta = beta;
    model.yloo = yloo;

    return model;
}
