function [model, loss_track, yloo] = train_multiple(hp, K, src_pred, y, newy)
%
% Trains a target classfier for N+1 classes, given the source classifier for N classes and the training set
% containing examples from all the N+1 classes. The new, N+1-th class is trained by leveraging upon the provided N class classifier.
%
% [model, loss_track, yloo] = train_multiple(hp, K, src_pred, y, newy)
%
% Input:
% hp.lambda -- KRLS regularization parameter
% hp.tau -- MULTIpLE regularization parameter (beta constraint)
% hp.iter -- MULTIpLE max subgradient descent iterations (default 10^4)
% K -- gram matrix of training data of size m
% src_pred -- source predictions on training data (# examples X # sources)
% y -- labels in {1,2,3, ..., N+1}^m
% newy -- label of new class. Should be present in y.
%
% Output:
% model.A -- KRLS parameters
% model.b -- KRLS bias term
% model.hp -- hyperparameters used during training
% model.beta -- source weights found by MULTIpLE
% model.new_class_ix -- 0/1 mask of a new class in the list of classes
% model.ova_code -- OVA list of classes
% loss_track -- losses recorded using subgrad. descent
% yloo -- matrix of leave-one-out predictions at the last step of subgradient descent
%
% The algorithm is described in
%
% "From N to N+ 1: Multiclass Transfer Incremental Learning."
% Kuzborskij, Ilja, Francesco Orabona, and Barbara Caputo.
% Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference on. IEEE, 2013.
%
% Contact the author: ilja.kuzborskij [at] idiap.ch
%                     http://idiap.ch/~ikuzbor
%
% ----------------------------------------------------------------------------------
%
% Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
% Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
%
% This file is part of MULTIpLE.
%
% MULTIpLE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License version 3 as
% published by the Free Software Foundation.
%
% MULTIpLE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with MULTIpLE. If not, see <http://www.gnu.org/licenses/>.
%
    if ~isfield(hp, 'iter')
        hp.iter = 10000; end
    loss_track = zeros(1, hp.iter);

    [yova, code] = ova(y);

    [model.A, model.b, model.beta, yloo] = train_multiple_mex(hp, K, yova, src_pred, find(code==newy));
    model.hp = hp;
    model.new_class_ix = (code==newy);
    model.ova_code = code;
    loss_track = [];
end
