/*
 Trains a target classfier for N+1 classes given the predictions of the source classifier for the N classes and the training set
 containing examples from all the N+1 classes. The new, N+1-th class is trained by leveraging upon the provided N class classifier predictions.

 [A, b, beta, yloo] = train_multiple_mex(hp, K, Y, src_pred, newy);

 Input:
 hp.lambda -- KRLS regularization parameter
 hp.tau -- MULTIpLE regularization parameter (beta constraint)
 hp.iter -- MULTIpLE max subgradient descent iterations (default 10^4)
 K -- gram matrix of training data of size m
 Y -- training label matrix by One-Versus-All coding (# examples X # classes)
 src_pred -- source predictions on training data (# examples X # sources)
 newy -- label of new class. Should be present in y.

 Output:
 A -- KRLS parameters
 b -- KRLS bias term
 hp -- hyperparameters used during training
 beta -- source weights found by MULTIpLE
 yloo -- matrix of leave-one-out predictions at the last step of subgradient descent

 The algorithm is described in

 "From N to N+ 1: Multiclass Transfer Incremental Learning."
 Kuzborskij, Ilja, Francesco Orabona, and Barbara Caputo.
 Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference on. IEEE, 2013.

 Contact the author: ilja.kuzborskij [at] idiap.ch
                     http://idiap.ch/~ikuzbor

 ----------------------------------------------------------------------------------

 Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
 Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>

 This file is part of MULTIpLE.

 MULTIpLE is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 3 as
 published by the Free Software Foundation.

 MULTIpLE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with MULTIpLE. If not, see <http://www.gnu.org/licenses/>.
*/

#include <mex.h>
#include <math.h>
#include <string.h>

#include <Eigen/Core>
#include <Eigen/Dense>
#include "train_multiple.hpp"

#include <iostream>

#define GETELEMENT(mat, row, col, colsize)   mat[col*colsize+row]
#define GETPTR(mat, idx, size)           (mat+idx*size)

char* expect_string(const mxArray* structure, char* key) {
    mxArray *field;
    field = mxGetField(structure, 0, key);
    if(field == NULL || !mxIsChar(field) || mxGetM(field) != 1) {
        mexErrMsgIdAndTxt("MATLAB:train_multiple_mex",
                          "Expected to read a string from the struct.");
    }
    return mxArrayToString(field);
}

double expect_double(const mxArray* structure, char* key) {
    mxArray *field;
    field = mxGetField(structure, 0, key);
    if(field == NULL || !mxIsNumeric(field) || mxGetM(field) != 1 || mxGetN(field) != 1) {
        mexErrMsgIdAndTxt("MATLAB:train_multiple_mex:expectDouble",
                          "Expecting to read a double from the struct.");
    }
    return *((double*) mxGetData(field));
}

MatrixXd double_to_eigen_mat(double *X, mwSignedIndex m, mwSignedIndex n) {
    MatrixXd M(m, n);
    for (mwSignedIndex i=0; i < m; ++i) {
        for (mwSignedIndex j=0; j < n; ++j) {
            M(i,j) = GETELEMENT(X, i, j, m);
        }
    }
    return M;
}

VectorXd double_to_eigen_vec(double *y, mwSignedIndex m) {
    VectorXd v(m);

    for (mwSignedIndex i=0; i < m; ++i) {
        v(i) = y[i];
    }
    return v;
}

void eigen_vec_to_double(double *out, VectorXd &v) {
    for (mwSignedIndex i=0; i < v.size(); ++i) {
        out[i] = v(i);
    }
}

void eigen_mat_to_double(double *out, MatrixXd &m) {
    for (mwSignedIndex i=0; i < m.rows(); ++i) {
        for (mwSignedIndex j=0; j < m.cols(); ++j) {
            GETELEMENT(out, i, j, m.rows()) = m(i,j);
        }
    }
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    const mxArray* config;

    /* Check I/O number */
    if(nlhs!=4) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:lhs",
			  "This function provides 4 output matrices.");
    }
    if(nrhs!=5) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:rhs",
			  "This function requires configuration struct, three matrices and one integer.");
    }

    if(!mxIsStruct(prhs[0])) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:rhs",
			  "This function requires the first parameter to be a configuration struct.");
    }

    size_t K_rows = (mwSignedIndex) mxGetM(prhs[1]);
    size_t K_cols = (mwSignedIndex) mxGetN(prhs[1]);
    MatrixXd K = double_to_eigen_mat(mxGetPr(prhs[1]), K_rows, K_cols);

    size_t Y_rows = (mwSignedIndex) mxGetM(prhs[2]);
    size_t Y_cols = (mwSignedIndex) mxGetN(prhs[2]);
    MatrixXd Y = double_to_eigen_mat(mxGetPr(prhs[2]), Y_rows, Y_cols);

    size_t src_pred_rows = (mwSignedIndex) mxGetM(prhs[3]);
    size_t src_pred_cols = (mwSignedIndex) mxGetN(prhs[3]);
    MatrixXd src_pred = double_to_eigen_mat(mxGetPr(prhs[3]), src_pred_rows, src_pred_cols);

    size_t new_class_pos = mxGetScalar(prhs[4]);

    config = prhs[0];

    size_t iterations = size_t(expect_double(config, "iter"));
    double lambda = expect_double(config, "lambda");
    double tau = expect_double(config, "tau");

    MultipleModel model = train_multiple(K, Y, src_pred, new_class_pos-1, iterations, lambda, tau);

    plhs[0] = mxCreateDoubleMatrix(model.A.rows(), model.A.cols(), mxREAL);
    eigen_mat_to_double(mxGetPr(plhs[0]), model.A);

    plhs[1] = mxCreateDoubleMatrix(model.b.rows(), model.b.cols(), mxREAL);
    eigen_mat_to_double(mxGetPr(plhs[1]), model.b);

    plhs[2] = mxCreateDoubleMatrix(model.beta.rows(), model.beta.cols(), mxREAL);
    eigen_vec_to_double(mxGetPr(plhs[2]), model.beta);

    plhs[3] = mxCreateDoubleMatrix(model.yloo.rows(), model.yloo.cols(), mxREAL);
    eigen_mat_to_double(mxGetPr(plhs[3]), model.yloo);

    // vec vecbeta = train_fr_tikhonov(matX, vecy, errtol, lambda);

    // plhs[0] = mxCreateDoubleMatrix(d, 1, mxREAL);
    // beta = mxGetPr(plhs[0]);
    // memset((void*) beta, 0, d*sizeof(mxREAL));

    // ublas_vec_to_double(beta, vecbeta);

    return;
}
